"""Hooks can be used to change default behaviour of parameter, ES or model setup."""
import logging
from inspect import signature
from copy import deepcopy
from dataclasses import dataclass
from enum import Enum
from typing import Any, Callable, Union


from django_oemof import settings


# pylint:disable=R0903
class AllScenarios:
    """Used to apply hook to all scenarios"""

    def __str__(self):
        return "ALL_SCENARIOS"


ALL_SCENARIOS = AllScenarios()

DEFAULT_PARAMETERS = ["scenario", "data"]


class HookType(Enum):
    """Hook types - define where to apply hooks"""

    SETUP = ("setup", [])
    PARAMETER = ("parameter", [])
    ENERGYSYSTEM = ("energysystem", ["energysystem"])
    MODEL = ("model", ["model"])
    POSTPROCESSING = ("postprocessing", ["model", "meta"])

    def __init__(self, label, parameters):
        self.label = label
        self.parameters = parameters


@dataclass
class Hook:
    """Hook class is used to set up a hook for specific scenario"""

    scenario: Union[str, AllScenarios]
    function: Callable

    def __str__(self):
        return f"<Hook '{self.function.__name__}' @{self.scenario}>"


def register_hook(hook_type: HookType, hook: Hook):
    """Registers hook depending on hook type"""
    sig = signature(hook.function)
    function_parameters = sig.parameters.keys()
    if any(parameter not in function_parameters for parameter in hook_type.parameters + DEFAULT_PARAMETERS):
        error_msg = f"Hook function '{hook.function.__name__}' misses parameters. Needed parameters: {hook_type.parameters + DEFAULT_PARAMETERS}, given: {list(function_parameters)}"
        raise KeyError(error_msg)
    if list(function_parameters)[:2] != DEFAULT_PARAMETERS:
        error_msg = f"Hook function '{hook.function.__name__}' must start with parameters {DEFAULT_PARAMETERS}, instead it starts with: {list(function_parameters)[:2]}"
        raise KeyError(error_msg)
    settings.HOOKS[hook_type].append(hook)


def apply_hooks(hook_type: HookType, scenario: str, data: Any, **kwargs) -> dict:
    """Applies hooks for a given hook type and scenario"""
    hooked_data = deepcopy(data)
    if hook_type == HookType.POSTPROCESSING:
        meta = deepcopy(kwargs["meta"])

    for hook in settings.HOOKS[hook_type]:
        if hook.scenario != scenario and hook.scenario is not ALL_SCENARIOS:
            continue
        logging.info(f"Applying {hook}")
        if hook_type == HookType.ENERGYSYSTEM:
            hook.function(scenario, hooked_data, energysystem=kwargs["energysystem"])
        elif hook_type == HookType.MODEL:
            hook.function(scenario, hooked_data, model=kwargs["model"])
        elif hook_type == HookType.POSTPROCESSING:
            meta = hook.function(scenario, hooked_data, meta=meta, model=kwargs["model"])
        else:
            hooked_data = hook.function(scenario, hooked_data)

    if hook_type == HookType.POSTPROCESSING:
        return meta
    return hooked_data



"""Config for django app"""

from django.apps import AppConfig


class DjangoOemofConfig(AppConfig):
    """Config for django-oemof app"""

    default_auto_field = "django.db.models.BigAutoField"
    name = "django_oemof"



# Generated by Django 5.0.9 on 2025-05-21 13:24

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_oemof", "0005_alter_result_name"),
    ]

    operations = [
        migrations.AddField(
            model_name="oemofdataset",
            name="meta_results",
            field=models.JSONField(default=dict),
            preserve_default=False,
        ),
    ]



"""Views for django_oemof"""
import json
import logging

from celery.result import AsyncResult
from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import APIView

from django.views.generic import TemplateView

from django_oemof import hooks, results, settings, simulation, models


class SimulateEnergysystem(APIView):
    """View to build and simulate Oemof energysystem from datapackage"""

    @staticmethod
    def get(request):
        """
        Checks simulation run using celery task ID

        Parameters
        ----------
        request
            Holding celery task ID

        Returns
        -------
        Response
            holding simulation ID if simulation is ready, otherwise simulation ID is None
        """
        task_id = request.GET["task_id"]
        task = AsyncResult(task_id)
        if task.ready():
            if task.failed():
                logging.info(f"Task #{task.task_id} failed: {task.result.args[0]}")
                return Response({"msg": f"Simulation error: {task.result.args[0]}"}, status=status.HTTP_400_BAD_REQUEST)
            logging.info(f"Task #{task.task_id} finished.")
            try:
                simulation_id = task.get()
            except:  # noqa: E722
                return Response({"msg": "Simulation error"}, status=status.HTTP_500_INTERNAL_SERVER_ERROR)
            if simulation_id is None:
                return Response({"msg": "Simulation is infeasible"}, status=status.HTTP_400_BAD_REQUEST)
            return Response({"simulation_id": simulation_id})
        return Response({"simulation_id": None})

    @staticmethod
    def post(request):
        """
        Simulates ES given by scenario and parameters

        Parameters
        ----------
        request
            Request holding scenario and parameters as JSON

        Returns
        -------
        Response
            holding celery task ID
        """
        scenario = request.POST["scenario"]
        parameters = hooks.apply_hooks(
            hook_type=hooks.HookType.SETUP, scenario=scenario, data=request.POST.dict()
        )
        task = simulation.simulate_scenario.delay(scenario, parameters)
        logging.info(f"Started simulation task #{task.task_id}.")
        return Response({"task_id": task.task_id})


class TerminateSimulationView(APIView):
    """View to terminate Oemof simulation run"""

    @staticmethod
    def post(request):
        """
        Delete task for given task ID

        Parameters
        ----------
        request
            Holding celery task ID

        Returns
        -------
        Response
            whether task deletion has been successful
        """
        task_id = request.POST["task_id"]
        task = AsyncResult(task_id)
        task.revoke(terminate=True)
        logging.info(f"Terminated task #{task_id}.")
        return Response()


class CalculateResults(APIView):
    """View calculate results from oemof simulation"""

    @staticmethod
    def get(request):
        """
        Calculates results for given scenario (with parameters)

        Parameters
        ----------
        request
            Request

        Returns
        -------
        Response
        """
        simulation_id = request.GET["simulation_id"]
        calculations = request.GET.getlist("calculations")
        calculated_results = results.get_results(simulation_id, calculations)
        return Response(calculated_results)


class FlowsView(TemplateView):
    template_name = "django_oemof/flows.html"

    def get_context_data(self, **kwargs):
        simulation_id = self.request.GET["simulation_id"]
        # Additional storages with different naming can be excluded via URl parameter "storages"
        storages = self.request.GET.getlist("storages") + ["storage", "battery"]
        try:
            sim = models.Simulation.objects.get(pk=simulation_id)  # pylint: disable=E1101
        except models.Simulation.DoesNotExist:  # pylint: disable=E1101
            # pylint: disable=W0707
            raise simulation.SimulationError(f"Simulation with ID#{simulation_id} not present in database.")

        _, results = sim.dataset.restore_results()
        links = [{"source": source, "target": target, "value": data["sequences"]["flow"].sum()} for (source, target), data in results.items() if "flow" in data["sequences"]]
        # Filter storages in order to avoid cycles (not possible with eCharts Sankeys) and zeros to avoid thin lines
        links = [link for link in links if all(storage not in link["target"] for storage in storages) and link["value"] > 0]
        names_raw = {link["source"] for link in links} | {link["target"] for link in links}
        names = [{"name": name} for name in names_raw]
        return {
            "results": results,
            "simulation_id": simulation_id,
            "chart_data": {
                "names": names,
                "links": links,
            },
        }



"""Tests for hooks"""

from django.test import TestCase

from django_oemof import hooks


class TestHooks(TestCase):
    """Testing hooks"""

    def test_parameter_hook(self):
        """Testing multiple hooks but for different scenarios"""

        def parameter_hook_1(scenario, data, request):
            return {key: f"{value}_1" for key, value in data.items()}

        def parameter_hook_2(scenario, data, request):
            return {key: f"{value}_2" for key, value in data.items()}

        def parameter_hook_all(scenario, data, request):
            return {key: f"{value}_all" for key, value in data.items()}

        hook1 = hooks.Hook(scenario="1", function=parameter_hook_1)
        hook2 = hooks.Hook(scenario="2", function=parameter_hook_2)
        hook_all = hooks.Hook(scenario=hooks.ALL_SCENARIOS, function=parameter_hook_all)
        hooks.register_hook(hooks.HookType.PARAMETER, hook1)
        hooks.register_hook(hooks.HookType.PARAMETER, hook2)
        hooks.register_hook(hooks.HookType.PARAMETER, hook_all)

        original_parameters = {"a": "1", "b": "2"}
        parameters_from_hook1 = hooks.apply_hooks(hooks.HookType.PARAMETER, scenario="1", data=original_parameters)
        assert parameters_from_hook1 == {"a": "1_1_all", "b": "2_1_all"}
        parameters_from_hook2 = hooks.apply_hooks(hooks.HookType.PARAMETER, scenario="2", data=original_parameters)
        assert parameters_from_hook2 == {"a": "1_2_all", "b": "2_2_all"}



"""Module to calculate oemof results"""

import inspect
from typing import Union, Type, Dict
import pandas

from oemof.tabular.postprocessing import core, calculations as standard_calculations

from . import simulation, models


CALCULATIONS: Dict[str, Union[Type[core.Calculation], core.ParametrizedCalculation]] = {
    core.get_dependency_name(member): member
    for (name, member) in inspect.getmembers(standard_calculations)
    if inspect.isclass(member) and not inspect.isabstract(member) and issubclass(member, core.Calculation)
}


def register_calculation(*calculations: Union[Type[core.Calculation], core.ParametrizedCalculation]):
    """
    Custom calculations have to be registered first, in order to use them via API
    """
    for calculation in calculations:
        CALCULATIONS[core.get_dependency_name(calculation)] = calculation


def get_results(
    simulation_id: int,
    calculations: Union[
        list[Union[str, Type[core.Calculation], core.ParametrizedCalculation]],
        dict[str, Union[str, Type[core.Calculation], core.ParametrizedCalculation]],
    ],
) -> dict[str, Union[pandas.Series, pandas.DataFrame]]:
    """
    Tries to load results from database.
    If result is not found, simulation data is loaded from db or simulated (if not in DB yet)
    and results are calculated.

    Parameters
    ----------
    simulation_id : int
        ID if simulation
    calculations : Union[
            list[Union[str, Type[core.Calculation], core.ParametrizedCalculation]],
            dict[str, Union[str, Type[core.Calculation], core.ParametrizedCalculation]],
        ]
        Either list or dict of calculations (by name or class) which shall be calculated.
        If dict is used, dict-keys are used keys in returned dict.

    Returns
    -------
    dict
        Dict containing calculation name (if dict is used for calculations input, input keys are used instead of
        internally used calculation name) as key and calculation result as value.
    """
    try:
        sim = models.Simulation.objects.get(pk=simulation_id)  # pylint: disable=E1101
    except models.Simulation.DoesNotExist:  # pylint: disable=E1101
        # pylint: disable=W0707
        raise simulation.SimulationError(f"Simulation with ID#{simulation_id} not present in database.")

    calculations = {
        calculation if isinstance(calculation, str) else core.get_dependency_name(calculation): calculation
        for calculation in calculations
    } if isinstance(calculations, list) else calculations

    results = {}
    for calculation_result_name, calculation in calculations.items():
        calculation_name = calculation if isinstance(calculation, str) else core.get_dependency_name(calculation)
        try:
            calculation_result = sim.results.get(name=calculation_name)
        except models.Result.DoesNotExist:  # pylint: disable=E1101
            continue
        result = pandas.read_json(calculation_result.data, orient="table")
        results[calculation_result_name] = result.iloc[:, 0] if calculation_result.data_type == "series" else result

    if len(results) != len(calculations):
        calculator = core.Calculator(*sim.dataset.restore_results())
        for calculation_result_name, calculation in calculations.items():
            if calculation_result_name in results:
                continue
            calculation_cls = CALCULATIONS[calculation] if isinstance(calculation, str) else calculation
            if isinstance(calculation_cls, core.ParametrizedCalculation):
                parameters = calculation_cls.parameters or {}
                result = calculation_cls.calculation(calculator, **parameters).result
            else:
                result = calculation_cls(calculator).result
            calculation_name = calculation if isinstance(calculation, str) else core.get_dependency_name(calculation)
            models.Result(
                simulation=sim,
                name=calculation_name,
                data=result.to_json(orient="table"),
                data_type="series" if isinstance(result, pandas.Series) else "frame",
            ).save()
            results[calculation_result_name] = result
    return results



import frictionless
from django.core.management.base import BaseCommand

from django_oemof.models import Simulation, Result
from django_oemof.settings import OEMOF_STATIC_DIR


class Command(BaseCommand):
    help = "Adds static result. Static result must be present in folder `oemof_static` in MEDIA_ROOT."

    def add_arguments(self, parser):
        parser.add_argument("scenarios", nargs="*", type=str, default=None)

    def handle(self, *args, **options):
        scenarios = options["scenarios"] or [scenario.name for scenario in OEMOF_STATIC_DIR.iterdir()]
        for scenario in scenarios:
            if Simulation.objects.filter(scenario=scenario).exists():
                self.stdout.write(
                    self.style.NOTICE(f"Simulation for scenario '{scenario}' already exists. Skipping...")
                )
                continue
            package = frictionless.Package(OEMOF_STATIC_DIR / scenario / "datapackage.json")
            parameters = package.get("parameters")
            simulation = Simulation(scenario=scenario, parameters=parameters)
            simulation.save()
            for resource in package.resources:
                df = resource.to_pandas()
                result = Result(
                    simulation=simulation,
                    name=resource.name,
                    data_type="series" if len(df.columns) == 2 else "frame",
                    data=df.to_json(orient="table"),
                )
                result.save()
            self.stdout.write(self.style.SUCCESS(f"Successfully created results for scenario '{scenario}'"))



"""Simulation module"""

import json
import logging
from typing import Optional
from collections import namedtuple

# pylint: disable=W0611
import oemof.tabular.datapackage  # noqa
from celery import shared_task
from django.conf import settings
from . import settings as do_settings
from oemof import solph
from oemof.tabular.facades import TYPEMAP

from django_oemof import hooks, models


FlowAttribute = namedtuple("FlowAttribute", ("from_node", "to_node", "attribute", "value"))


class SimulationError(Exception):
    """Raised if simulation failed or simulation is not present"""


@shared_task
def simulate_scenario(scenario: str, parameters: dict, lp_file: Optional[str] = None) -> int:
    """
    Returns ID to oemof results from simulated/restored scenario

    Already stored scenarios are identified by scenario name and
    adapted parameters.

    Parameters
    ----------
    scenario: str
        Name of scenario (used to load related datapackage)
    parameters: dict
        Parameters which are adapted to ES before simulation
    lp_file: Optional[str]
        If set, LP file is stored under given path

    Returns
    -------
    int
        Simulation ID where results are stored
    """
    try:
        simulation = models.Simulation.objects.get(scenario=scenario, parameters=parameters)  # pylint: disable=E1101
        logging.info(f"Simulation for {scenario=} and {parameters=} already present as #{simulation.id}.")
    except models.Simulation.DoesNotExist:  # pylint: disable=E1101
        logging.info(f"Simulating energysystem for {scenario=} and {parameters=}.")
        oemof_datapackage = f"{settings.MEDIA_ROOT}/oemof/{scenario}/datapackage.json"
        energysystem = build_energysystem(oemof_datapackage)
        build_parameters = hooks.apply_hooks(hook_type=hooks.HookType.PARAMETER, scenario=scenario, data=parameters)
        energysystem = adapt_energysystem(energysystem, build_parameters)
        hooks.apply_hooks(hook_type=hooks.HookType.ENERGYSYSTEM, scenario=scenario, data=build_parameters, energysystem=energysystem)
        termination_condition, input_data, results_data, meta_results = simulate_energysystem(
            scenario, energysystem, build_parameters, lp_file
        )
        if termination_condition == "infeasible":
            logging.warning(f"Simulation run for {scenario=} and {parameters=} is infeasible.")
            return
        dataset = models.OemofDataset.store_results(input_data, results_data, meta_results)
        # pylint: disable=E1101
        try:
            # Add additional check in case simulation with same parameters has been run in parallel
            simulation = models.Simulation.objects.get(scenario=scenario, parameters=parameters)
            logging.info(
                f"Simulation results for {scenario=} and {parameters=} are stored already by other simulation run."
            )
        except models.Simulation.DoesNotExist:
            simulation = models.Simulation.objects.create(scenario=scenario, parameters=parameters, dataset=dataset)
            simulation.save()
            logging.info(f"Stored simulation results for {scenario=} and {parameters=} as #{simulation.id}")
    return simulation.id


def build_energysystem(oemof_datapackage: str):
    """
    Builds energysystem from datapackage

    Parameters
    ----------
    oemof_datapackage: str
        Path to oemof.tabular datapackage

    Returns
    -------
    energysystem: Energysystem build from datapacakge
    """
    logging.info(f"Building energysystem from datapackage at '{oemof_datapackage}'.")
    return solph.EnergySystem.from_datapackage(oemof_datapackage, typemap=TYPEMAP)


def adapt_energysystem(energysystem: solph.EnergySystem, parameters: dict):
    """
    Adapt parameters in ES

    This allows loading standard ES and changing specific parameters after build.

    Parameters
    ----------
    energysystem: EnergySystem
        Components in Energysystem are changed by given parameters

    parameters: dict
        Parameters which shall be adapted in ES

    Returns
    -------
    energysystem: Energysystem with changed parameters
    """
    parameters = parameters or {}
    logging.info(f"Adapting parameters in ES using {parameters=}.")

    for node_name, attributes in parameters.items():
        if node_name == "flow":
            logging.warning(
                f"This is deprecated. Flows are adapted using input_parameters and output_parameters instead."
            )
            continue
        if node_name not in [n.label for n in energysystem.nodes]:
            log_msg = f"Cannot adapt component '{node_name}', as it cannot be found in energysystem."
            logging.warning(log_msg)
            continue
        if not isinstance(attributes, dict):
            log_msg = (
                f"Cannot adapt attributes for component '{node_name}', as there is no dictionary. Skipping it." 
                """Try something like {"wind_onshore": {"capacity": 100}} instead."""
            )
            logging.warning(log_msg)
            continue

        node = next(n for n in energysystem.nodes if node_name == n.label)
        for attribute, value in attributes.items():
            if not hasattr(node, attribute):
                logging.warning(
                    f"Attribute '{attribute}' not found in component '{node_name}' in energysystem. "
                    "Adapting the attribute might have no effect."
                )
            setattr(node, attribute, value)
        node.update()
    return energysystem


def simulate_energysystem(scenario, energysystem, user_data: Optional[dict] = None, lp_file: Optional[str] = None):
    """
    Simulates ES, stores results to DB and returns simulation ID

    Parameters
    ----------
    scenario: str
        Name of current scenario (used to apply hooks)
    energysystem : EnergySystem
        Built energysystem to be solved
    user_data: Optional[dict]
        Optional user data to be passed to hooks
    lp_file: Optional[str]
        If set, LP file is stored under given path

    Returns
    -------
    results : tuple(bool, dict, dict, dict, dict)
        Simulation termination condition, input, results, meta results and custom results
    """
    logging.info(f"Building model for {scenario=}.")
    model = solph.Model(energysystem)
    hooks.apply_hooks(hook_type=hooks.HookType.MODEL, scenario=scenario, data=user_data, model=model)
    logging.info(f"Starting simulation for {scenario=}.")
    model_results = model.solve(
        solver="cbc", cmdline_options={"mipgap": "0.1", "seconds": do_settings.DJANGO_OEMOF_TIMELIMIT}
    )
    if lp_file:
        model.write(lp_file, io_options={"symbolic_solver_labels": True})
    logging.info(f"Simulation for {scenario=} finished.")

    input_data = solph.processing.parameter_as_dict(
        energysystem,
        exclude_attrs=["bus", "from_bus", "to_bus", "from_node", "to_node"],
    )
    results_data = solph.processing.results(model)
    # Clean-up meta results by dumping and loading and neglecting non-serializable key/values
    meta_results = json.loads(
        json.dumps(solph.processing.meta_results(model), skipkeys=True, default=lambda x: "Not serializable")
    )
    meta_results = hooks.apply_hooks(
        hook_type=hooks.HookType.POSTPROCESSING, scenario=scenario, data=user_data, model=model, meta=meta_results
    )

    return (
        model_results.solver.termination_condition,
        *map(solph.processing.convert_keys_to_strings, (input_data, results_data)),
        meta_results,
    )



"""Settings for django_oemof"""

import environ
import pathlib
from collections import defaultdict

from django.conf import settings

env = environ.Env()

OEMOF_DIR = pathlib.Path(settings.MEDIA_ROOT) / getattr(settings, "DJANGO_OEMOF_DIR", "oemof")
OEMOF_STATIC_DIR = pathlib.Path(settings.MEDIA_ROOT) / getattr(settings, "DJANGO_OEMOF_STATIC_DIR", "oemof_static")

HOOKS = defaultdict(list)

DJANGO_OEMOF_IGNORE_SIMULATION_PARAMETERS = env.list("DJANGO_OEMOF_IGNORE_SIMULATION_PARAMETERS", default=[])
DJANGO_OEMOF_TIMELIMIT = env.int("DJANGO_OEMOF_TIMELIMIT", default=600)



"""Postprocessing test module"""

from django.test import SimpleTestCase
from oemof.tabular.postprocessing import calculations, core

from django_oemof import results as dor
from django_oemof import simulation, models

OEMOF_DATAPACKAGE = "test_scenario"


class TestCalculation(core.Calculation):
    """Example calculation for testing"""

    name = "test"
    depends_on = {"summed_flows": calculations.AggregatedFlows}

    def calculate_result(self):
        return self.dependency("summed_flows") * 5


class PostprocessingTest(SimpleTestCase):
    """
    SimpleTestCase in combination with `database` and overrideen `tearDown` is used to preserve data in database,
    as `--keepdb` only keeps tables, not records
    """

    databases = "__all__"  # This is used to preserve data in database (--keepdb only keeps tables, not records)

    def test_postprocessing(self):
        """Tests postprocessing of oemof simulation"""
        simulation_id = simulation.simulate_scenario(OEMOF_DATAPACKAGE, {})
        sim = models.Simulation.objects.get(pk=simulation_id)
        calculator = core.Calculator(*sim.dataset.restore_results())
        total_system_costs = calculations.TotalSystemCosts(calculator)
        print(total_system_costs.result)
        assert not total_system_costs.result.empty

    def test_results(self):
        """Tests registration of custom calculation"""
        # Register calculation test:
        dor.register_calculation(TestCalculation)
        simulation_id = simulation.simulate_scenario(OEMOF_DATAPACKAGE, {})
        results = dor.get_results(simulation_id, calculations=["test"])
        assert "test" in results
        assert not results["test"].empty

    def test_results_with_calculation_cls(self):
        """Tests registration of custom calculation"""
        # Register calculation test:
        dor.register_calculation(TestCalculation)
        simulation_id = simulation.simulate_scenario(OEMOF_DATAPACKAGE, {})
        results = dor.get_results(simulation_id, calculations=[TestCalculation])
        assert "test" in results
        assert not results["test"].empty

    def tearDown(self) -> None:
        # Must stay empty in order to keep simulation data in test db
        pass

    @classmethod
    def tearDownClass(cls):
        # Must stay empty in order to keep simulation data in test db
        pass



# Generated by Django 3.2.19 on 2023-06-16 09:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_oemof', '0004_alter_simulation_dataset'),
    ]

    operations = [
        migrations.AlterField(
            model_name='result',
            name='name',
            field=models.TextField(),
        ),
    ]



# Generated by Django 3.2.16 on 2022-11-11 09:29

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_oemof', '0002_simulation'),
    ]

    operations = [
        migrations.CreateModel(
            name='Result',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('data_type', models.CharField(choices=[('series', 'series'), ('frame', 'frame')], max_length=9)),
                ('data', models.JSONField()),
                ('simulation', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='results', to='django_oemof.simulation')),
            ],
        ),
    ]



"""Tests storing and restoring of oemof results in DB"""

import pathlib

# pylint: disable=W0611
import oemof.tabular.datapackage  # noqa
from django.test import TransactionTestCase
from oemof import solph

from django_oemof import models, simulation

OEMOF_DATAPACKAGE = pathlib.Path(__file__).parent / "test_data" / "dispatch" / "datapackage.json"


class OemofDBTest(TransactionTestCase):
    """Test case for (re-)storing oemof results in DB"""

    def test_store_and_restore(self):
        """Stores and restores results from simple oemof example"""
        energysystem = simulation.build_energysystem(str(OEMOF_DATAPACKAGE))
        input_data, results_data = simulation.simulate_energysystem(energysystem)
        input_data = solph.processing.convert_keys_to_strings(input_data)
        results_data = solph.processing.convert_keys_to_strings(results_data)

        result_id = models.OemofDataset.store_results(input_data, results_data)
        assert result_id == 1
        dataset = models.OemofDataset.objects.get(pk=result_id)  # pylint: disable=E1101
        restored_input, restored_results = dataset.restore_results()
        assert len(input_data) == len(restored_input)
        assert len(input_data) == len(restored_input)
        assert len(results_data) == len(restored_results)
        assert len(results_data) == len(restored_results)



# Generated by Django 3.2.16 on 2023-01-06 03:56

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_oemof', '0003_result'),
    ]

    operations = [
        migrations.AlterField(
            model_name='simulation',
            name='dataset',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='django_oemof.oemofdataset'),
        ),
    ]



import django
import environ
from django.conf import settings
from django.core.management import execute_from_command_line


def init_django(installed_apps=None):
    if settings.configured:
        return

    env = environ.Env()
    env.read_env(".env")

    installed_apps = installed_apps or []

    settings.configure(
        INSTALLED_APPS=[
            "django_oemof",
        ]
        + installed_apps,
        DATABASES={"default": env.db("DATABASE_URL")},
        MEDIA_ROOT=env.path("MEDIA_ROOT", default="media"),
    )
    django.setup()


if __name__ == "__main__":
    init_django()
    execute_from_command_line()



"""Module holds models to store and restore oemof results in/from DB"""

import pandas
from django.contrib.postgres.fields import ArrayField
from django.db import models
from oemof.network.network import Node
from oemof.solph.processing import convert_keys_to_strings


class Simulation(models.Model):
    """Holds information about simulation input parameters and related oemof results"""

    scenario = models.CharField(max_length=255)
    parameters = models.JSONField()
    dataset = models.ForeignKey("OemofDataset", on_delete=models.CASCADE, null=True)


class OemofDataset(models.Model):
    """Holds inputs and results of an oemof solph optimization"""

    input = models.ForeignKey("OemofData", on_delete=models.CASCADE, related_name="data_input")  # noqa: A003
    result = models.ForeignKey("OemofData", on_delete=models.CASCADE, related_name="data_result")
    meta_results = models.JSONField()

    # pylint: disable=R0914
    @classmethod
    def store_results(cls, input_data, result_data, meta_results):
        """
        Stores inputs and results from oemof into DB

        For each in entry in scalars and sequences of both input- and result-data
        an OemofScalar or OemofSequence is build and connected to an OemofData
        object representing either input or result data. At last, both OemofData
        objects are connected to OemofInputResult object and resulting index is
        returned.

        Parameters
        ----------
        input_data: dict
            Output of oemof.solph.processing.param_results with nodes as str
            (use oemof.solph.processing.convert_keys_to_str if necessary)
        result_data: dict
            Output of oemof.solph.processing.param_results with nodes as str
            (use oemof.solph.processing.convert_keys_to_str if necessary)
        meta_results: dict
            Output of oemof.solph.processing.meta_results and additonal data from custom postprocessing hooks

        Returns
        -------
        OemofDataset: Instance of created OemofDataset
        """
        if not isinstance(next(iter(input_data)), str):
            input_data = convert_keys_to_strings(input_data)
        if not isinstance(next(iter(result_data)), str):
            result_data = convert_keys_to_strings(result_data)
        oemof_dataset = OemofDataset()
        oemof_dataset.meta_results = meta_results
        for input_result_attr, data in (("input", input_data), ("result", result_data)):
            scalars = []
            sequences = []
            for (from_node, to_node), sc_sq_dict in data.items():
                if "scalars" in sc_sq_dict:
                    for key, value in sc_sq_dict["scalars"].items():
                        if isinstance(value, Node):
                            continue
                        scalar = OemofScalar(
                            from_node=from_node, to_node=to_node, attribute=key, value=value, type=type(value).__name__
                        )
                        scalar.save()
                        scalars.append(scalar)
                for key, series in sc_sq_dict["sequences"].items():
                    list_type = "list"
                    if isinstance(series, pandas.Series):
                        series = series.values.tolist()
                        list_type = "series"
                    sequence = OemofSequence(
                        from_node=from_node, to_node=to_node, attribute=key, value=series, type=list_type
                    )
                    sequence.save()
                    sequences.append(sequence)
            oemof_data = OemofData()
            oemof_data.save()
            oemof_data.scalars.set(scalars)  # pylint: disable=E1101
            oemof_data.sequences.set(sequences)  # pylint: disable=E1101
            setattr(oemof_dataset, input_result_attr, oemof_data)
        oemof_dataset.save()
        return oemof_dataset

    def restore_results(self):
        """
        Restores input and result data from OemofDataset

        Returns
        -------
        (dict, dict):
            Restored input- and result-data
        """

        def type_conversion(value_str, value_type):
            if value_type == "str":
                return value_str
            if value_type in ("float", "float64"):
                return float(value_str)
            if value_type in ("int", "int64"):
                return int(value_str)
            if value_type == "bool":
                return bool(value_str)
            raise TypeError('Unknown conversion type "' + value_type + '"')

        input_data = {}
        result_data = {}
        for input_result_attr, data in (("input", input_data), ("result", result_data)):
            ir_attr = getattr(self, input_result_attr)
            for scalar in ir_attr.scalars.all():  # pylint: disable=E1101
                nodes = (scalar.from_node, scalar.to_node)
                if nodes not in data:
                    data[nodes] = {"scalars": {}, "sequences": {}}
                data[nodes]["scalars"][scalar.attribute] = type_conversion(scalar.value, scalar.type)
            for sequence in ir_attr.sequences.all():  # pylint: disable=E1101
                nodes = (sequence.from_node, sequence.to_node)
                if nodes not in data:
                    data[nodes] = {"scalars": {}, "sequences": {}}
                if sequence.type == "series":
                    series = pandas.Series(sequence.value)
                else:
                    series = sequence.value
                data[nodes]["sequences"][sequence.attribute] = series
        return input_data, result_data


class OemofData(models.Model):
    """Model to hold input or results data from oemof"""

    scalars = models.ManyToManyField("OemofScalar")
    sequences = models.ManyToManyField("OemofSequence")


class OemofScalar(models.Model):
    """Represents a scalar from oemof results"""

    from_node = models.CharField(max_length=255)
    to_node = models.CharField(max_length=255)
    attribute = models.CharField(max_length=255)
    value = models.CharField(max_length=255)
    type = models.CharField(max_length=255)  # noqa: A003


class OemofSequence(models.Model):
    """Represents a sequence from oemof results"""

    from_node = models.CharField(max_length=255)
    to_node = models.CharField(max_length=255, null=True)
    attribute = models.CharField(max_length=255)
    value = ArrayField(base_field=models.FloatField())
    type = models.CharField(max_length=255)  # noqa: A003


class Result(models.Model):
    """Model to store results from postprocessing"""

    simulation = models.ForeignKey("Simulation", related_name="results", on_delete=models.CASCADE)
    name = models.TextField()
    data_type = models.CharField(max_length=9, choices=(("series", "series"), ("frame", "frame")))
    data = models.JSONField()



# Generated by Django 3.2.15 on 2022-09-23 12:04

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_oemof', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Simulation',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('scenario', models.CharField(max_length=255)),
                ('parameters', models.JSONField()),
                ('dataset', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_oemof.oemofdataset')),
            ],
        ),
    ]



"""Tests to parametrize oemof energysystems before simulation start"""

import copy
import pathlib
import time
import unittest

# pylint: disable=W0611
import oemof.tabular.datapackage  # noqa
import pytest
from django.test import TransactionTestCase

from django_oemof import simulation

OEMOF_DATAPACKAGE = pathlib.Path(__file__).parent / "test_data" / "oemof" / "dispatch" / "datapackage.json"
OEMOF_BIG_DATAPACKAGE = pathlib.Path(__file__).parent / "test_data" / "oemof" / "test_scenario" / "datapackage.json"


class OemofParameterTest(TransactionTestCase):
    """Test case for paramtrizing oemof ES after initialization"""

    def test_without_parameterization(self):
        """Build ES from datapacakge and simulate as reference to 'test_with_parameterization'"""
        energysystem = simulation.build_energysystem(str(OEMOF_DATAPACKAGE))
        _, results_data = simulation.simulate_energysystem(energysystem)
        assert list(results_data.values())[6]["sequences"].flow[0] == pytest.approx(7.37660)
        assert list(results_data.values())[6]["sequences"].flow[1] == pytest.approx(9.20905)
        assert list(results_data.values())[6]["sequences"].flow[2] == pytest.approx(11.19685)

    def test_with_parameterization(self):
        """Build ES from datapacakge and change parameter afterwards before simulation"""
        parameters = {"wind": {"capacity": 5}}
        energysystem = simulation.build_energysystem(str(OEMOF_DATAPACKAGE))
        energysystem = simulation.adapt_energysystem(energysystem, parameters)
        _, results_data = simulation.simulate_energysystem(energysystem)
        assert list(results_data.values())[6]["sequences"].flow[0] == pytest.approx(7.37660 / 10)
        assert list(results_data.values())[6]["sequences"].flow[1] == pytest.approx(9.20905 / 10)
        assert list(results_data.values())[6]["sequences"].flow[2] == pytest.approx(11.19685 / 10)


class OemofBuildTest(TransactionTestCase):
    """Test case for copying existing ES"""

    def test_build_large_es(self):
        """Test building a large ES"""
        start = time.time()
        energysystem = simulation.build_energysystem(str(OEMOF_BIG_DATAPACKAGE))
        print("Time to build ES:", time.time() - start)  # ~5s
        start = time.time()
        simulation.simulate_energysystem(energysystem)
        print("Time to simulate ES:", time.time() - start)  # 28s

    @unittest.skip("Copying ES does not work!")
    def test_copying_es(self):
        """Test to copy ES"""
        energysystem = simulation.build_energysystem(str(OEMOF_DATAPACKAGE))
        energysystem_copy = copy.deepcopy(energysystem)
        parameters = {"wind": {"capacity": 5}}
        energysystem_changed = simulation.adapt_energysystem(energysystem_copy, parameters)
        _, results_data = simulation.simulate_energysystem(energysystem)
        _, results_data_changed = simulation.simulate_energysystem(energysystem_changed)

        assert list(results_data.values())[6]["sequences"].flow[0] == pytest.approx(7.37660)
        assert list(results_data_changed.values())[6]["sequences"].flow[0] == pytest.approx(7.37660 / 10)



# Generated by Django 3.2.15 on 2022-09-16 09:01

import django.contrib.postgres.fields
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='OemofData',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
        ),
        migrations.CreateModel(
            name='OemofScalar',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('from_node', models.CharField(max_length=255)),
                ('to_node', models.CharField(max_length=255)),
                ('attribute', models.CharField(max_length=255)),
                ('value', models.CharField(max_length=255)),
                ('type', models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name='OemofSequence',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('from_node', models.CharField(max_length=255)),
                ('to_node', models.CharField(max_length=255, null=True)),
                ('attribute', models.CharField(max_length=255)),
                ('value', django.contrib.postgres.fields.ArrayField(base_field=models.FloatField(), size=None)),
                ('type', models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name='OemofDataset',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('input', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='data_input', to='django_oemof.oemofdata')),
                ('result', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='data_result', to='django_oemof.oemofdata')),
            ],
        ),
        migrations.AddField(
            model_name='oemofdata',
            name='scalars',
            field=models.ManyToManyField(to='django_oemof.OemofScalar'),
        ),
        migrations.AddField(
            model_name='oemofdata',
            name='sequences',
            field=models.ManyToManyField(to='django_oemof.OemofSequence'),
        ),
    ]



"""Settings for testing django-oemof app"""

import pathlib
import environ

env = environ.Env()

DATABASES = {"default": env.db("DATABASE_URL")}

TEST_FOLDER = pathlib.Path(__file__).parent
TEST_PATH = "django_oemof.tests"

INSTALLED_APPS = (
    "django.contrib.contenttypes",
    "django.contrib.auth",
    "django.contrib.sites",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.admin.apps.SimpleAdminConfig",
    "django.contrib.staticfiles",
    "rest_framework",
    "django_oemof",
)

ROOT_URLCONF = ""  # tests override urlconf, but it still needs to be defined
MIDDLEWARE = (
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
)
TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]
MEDIA_ROOT = TEST_FOLDER / "test_data"



"""Urls for django-oemof package"""

from django.urls import include, path
from rest_framework.routers import DefaultRouter

from django_oemof import views

router = DefaultRouter()

# pylint:disable=C0103
app_name = "django_oemof"

urlpatterns = [
    path("", include(router.urls)),
    path("simulate", views.SimulateEnergysystem.as_view(), name="simulate"),
    path("terminate", views.TerminateSimulationView.as_view()),
    path("calculate", views.CalculateResults.as_view()),
    path("flows", views.FlowsView.as_view()),
]
